@@ -10,8 +10,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not, see: <http://www.gnu.org/licenses/>
  */
 
 /* ---------------------------- included header files ---------------------- */
@@ -87,20 +86,25 @@ void gravity_add_decoration(
 	return;
 }
 
-void get_relative_geometry(rectangle *rel_g, rectangle *abs_g)
+void get_relative_geometry(FvwmWindow *fw, rectangle *rel_g, rectangle *abs_g)
 {
-	rel_g->x = abs_g->x - Scr.Vx;
-	rel_g->y = abs_g->y - Scr.Vy;
+	struct monitor	*m = (fw && fw->m) ? fw->m : monitor_get_current();
+
+	rel_g->x = abs_g->x - m->virtual_scr.Vx;
+	rel_g->y = abs_g->y - m->virtual_scr.Vy;
 	rel_g->width = abs_g->width;
 	rel_g->height = abs_g->height;
 
 	return;
 }
 
-void get_absolute_geometry(rectangle *abs_g, rectangle *rel_g)
+void get_absolute_geometry(FvwmWindow *fw, rectangle *abs_g, rectangle *rel_g)
 {
-	abs_g->x = rel_g->x + Scr.Vx;
-	abs_g->y = rel_g->y + Scr.Vy;
+	/* FIXME - not sure this is correct. */
+	struct monitor	*m = (fw && fw->m) ? fw->m : monitor_get_current();
+
+	abs_g->x = rel_g->x + m->virtual_scr.Vx;
+	abs_g->y = rel_g->y + m->virtual_scr.Vy;
 	abs_g->width = rel_g->width;
 	abs_g->height = rel_g->height;
 
@@ -343,7 +347,18 @@ void get_shaded_geometry(
 	int big_height = big_g->height;
 	int d;
 
-	get_window_borders(fw, &b);
+	switch (SHADED_DIR(fw))
+	{
+	case DIR_SW:
+	case DIR_SE:
+	case DIR_NW:
+	case DIR_NE:
+		get_window_borders_no_title(fw, &b);
+		break;
+	default:
+		get_window_borders(fw, &b);
+		break;
+	}
 	*small_g = *big_g;
 	d = 0;
 	switch (SHADED_DIR(fw))
@@ -420,7 +435,7 @@ void get_unshaded_geometry(
 		{
 			*ret_g = fw->g.normal;
 		}
-		get_relative_geometry(ret_g, ret_g);
+		get_relative_geometry(fw, ret_g, ret_g);
 	}
 	else
 	{
@@ -438,7 +453,7 @@ void get_shaded_client_window_pos(
 
 	get_window_borders(fw, &b);
 	big_g = (IS_MAXIMIZED(fw)) ? fw->g.max : fw->g.normal;
-	get_relative_geometry(&big_g, &big_g);
+	get_relative_geometry(fw, &big_g, &big_g);
 	switch (SHADED_DIR(fw))
 	{
 	case DIR_S:
@@ -567,7 +582,7 @@ void get_client_geometry(
  * state */
 void update_relative_geometry(FvwmWindow *fw)
 {
-	get_relative_geometry(
+	get_relative_geometry(fw,
 		&fw->g.frame,
 		(IS_MAXIMIZED(fw)) ? &fw->g.max : &fw->g.normal);
 	if (IS_SHADED(fw))
@@ -584,12 +599,13 @@ void update_absolute_geometry(FvwmWindow *fw)
 {
 	rectangle *dest_g;
 	rectangle frame_g;
+	struct monitor	*m = (fw && fw->m) ? fw->m : monitor_get_current();
 
 	/* store orig values in absolute coords */
 	dest_g = (IS_MAXIMIZED(fw)) ? &fw->g.max : &fw->g.normal;
 	frame_g = *dest_g;
-	dest_g->x = fw->g.frame.x + Scr.Vx;
-	dest_g->y = fw->g.frame.y + Scr.Vy;
+	dest_g->x = fw->g.frame.x + m->virtual_scr.Vx;
+	dest_g->y = fw->g.frame.y + m->virtual_scr.Vy;
 	dest_g->width = fw->g.frame.width;
 	dest_g->height = fw->g.frame.height;
 	if (IS_SHADED(fw))
@@ -633,16 +649,18 @@ void maximize_adjust_offset(FvwmWindow *fw)
 	int off_y;
 	int dh;
 	int dw;
+	struct monitor	*m;
 
 	if (!IS_MAXIMIZED(fw))
 	{
 		/* otherwise we might corrupt the g.normal */
 		return;
 	}
+	m = fw->m;
 	off_x = fw->g.normal.x - fw->g.max.x - fw->g.max_offset.x;
 	off_y = fw->g.normal.y - fw->g.max.y - fw->g.max_offset.y;
-	dw = Scr.MyDisplayWidth;
-	dh = Scr.MyDisplayHeight;
+	dw = m->virtual_scr.MyDisplayWidth;
+	dh = m->virtual_scr.MyDisplayHeight;
 	if (off_x >= dw)
 	{
 		fw->g.normal.x -= (off_x / dw) * dw;
@@ -808,6 +826,7 @@ void constrain_size(
 	size_rect d;
 	size_rect old;
 	size_borders b;
+	struct monitor	*m = fw->m;
 
 	if (DO_DISABLE_CONSTRAIN_SIZE_FULLSCREEN(fw) == 1)
 	{
@@ -930,7 +949,7 @@ void constrain_size(
 		}
 		else if (
 			xmotion < 0 && e->xmotion.x_root >=
-			Scr.MyDisplayWidth - round_up.width)
+			m->virtual_scr.MyDisplayWidth - round_up.width)
 		{
 			d.width -= inc.width;
 		}
@@ -940,7 +959,7 @@ void constrain_size(
 		}
 		else if (
 			ymotion < 0 && e->xmotion.y_root >=
-			Scr.MyDisplayHeight - round_up.height)
+			m->virtual_scr.MyDisplayHeight - round_up.height)
 		{
 			d.height -= inc.height;
 		}
@@ -1334,17 +1353,21 @@ void resize_icon_title_height(FvwmWindow *fw, int dh)
 	return;
 }
 
-void get_page_offset_rectangle(
+void get_page_offset_rectangle(FvwmWindow *fw,
 	int *ret_page_x, int *ret_page_y, rectangle *r)
 {
-	int xoff = Scr.Vx % Scr.MyDisplayWidth;
-	int yoff = Scr.Vy % Scr.MyDisplayHeight;
+	struct monitor	*m = (fw && fw->m) ? fw->m : monitor_get_current();
+
+	/* FIXME: broadcast if global monitor in use. */
+
+	int xoff = m->virtual_scr.Vx % m->virtual_scr.MyDisplayWidth;
+	int yoff = m->virtual_scr.Vy % m->virtual_scr.MyDisplayHeight;
 
 	/* maximize on the page where the center of the window is */
 	*ret_page_x = truncate_to_multiple(
-		r->x + r->width / 2 + xoff, Scr.MyDisplayWidth) - xoff;
+		r->x + r->width / 2 + xoff, m->virtual_scr.MyDisplayWidth) - xoff;
 	*ret_page_y = truncate_to_multiple(
-		r->y + r->height / 2 + yoff, Scr.MyDisplayHeight) - yoff;
+		r->y + r->height / 2 + yoff, m->virtual_scr.MyDisplayHeight) - yoff;
 
 	return;
 }
@@ -1358,7 +1381,7 @@ void get_page_offset(
 	r.y = fw->g.frame.y;
 	r.width = fw->g.frame.width;
 	r.height = fw->g.frame.height;
-	get_page_offset_rectangle(ret_page_x, ret_page_y, &r);
+	get_page_offset_rectangle(fw, ret_page_x, ret_page_y, &r);
 
 	return;
 }
@@ -1366,7 +1389,7 @@ void get_page_offset(
 void get_page_offset_check_visible(
 	int *ret_page_x, int *ret_page_y, FvwmWindow *fw)
 {
-	if (IsRectangleOnThisPage(&fw->g.frame, fw->Desk))
+	if (IsRectangleOnThisPage(fw->m, &fw->g.frame, fw->Desk))
 	{
 		/* maximize on visible page if any part of the window is
 		 * visible */
