@@ -10,8 +10,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not, see: <http://www.gnu.org/licenses/>
  */
 
 #include "config.h"
@@ -156,7 +155,8 @@ Window find_frame_window (Window win, int *off_x, int *off_y)
   {
     if (!XQueryTree (theDisplay, win, &root, &parent, &junkw, &junki))
     {
-      fprintf(stderr, "%s: find_frame_window: XQueryTree failed.\n", MyName);
+      fvwm_debug(__func__, "%s: find_frame_window: XQueryTree failed.\n",
+                 MyName);
       return win;
     }
     if (junkw)
@@ -165,8 +165,9 @@ Window find_frame_window (Window win, int *off_x, int *off_y)
       break;
     if (!XGetWindowAttributes (theDisplay, win, &attr))
     {
-      fprintf(stderr, "%s: find_frame_window: XGetWindowAttrigutes failed.\n",
-	      MyName);
+      fvwm_debug(__func__,
+                 "%s: find_frame_window: XGetWindowAttrigutes failed.\n",
+                 MyName);
       return win;
     }
     ConsoleDebug (X11, "Adding (%d, %d) += (%d, %d)\n",
@@ -491,7 +492,7 @@ void xevent_loop (void)
 
     case MapNotify:
       ConsoleDebug (X11, "XEVENT: MapNotify\n");
-      if (globals.transient && !grab_state != HAVE_GRABBED) {
+      if (globals.transient) {
 	grab_pointer (man);
       }
       force_manager_redraw (man);
@@ -737,13 +738,18 @@ void X_init_manager (int man_id)
   man->geometry.x = man->managed_g.x;
   man->geometry.y = man->managed_g.y;
   if (man->geometry_str) {
-    int scr;
+    char *scr;
+    fscreen_scr_arg arg;
+    arg.mouse_ev = NULL;
 
     geometry_mask = FScreenParseGeometryWithScreen(
       man->geometry_str, &man->geometry.x, &man->geometry.y,
       &man->geometry.cols, &man->geometry.rows, &scr);
+
+    arg.name = scr;
+
     FScreenGetScrRect(
-      NULL, scr, &man->managed_g.x, &man->managed_g.y,
+      &arg, FSCREEN_BY_NAME, &man->managed_g.x, &man->managed_g.y,
       &man->managed_g.width, &man->managed_g.height);
 
     if (geometry_mask & XValue)
@@ -855,7 +861,6 @@ void create_manager_window (int man_id)
   int join_style = JoinRound;
   int i;
   WinManager *man;
-  fscreen_scr_t scr;
   ConsoleDebug (X11, "In create_manager_window\n");
 
   man = &globals.managers[man_id];
@@ -922,7 +927,6 @@ void create_manager_window (int man_id)
       &fscr, FSCREEN_XYPOS,
       &man->managed_g.x, &man->managed_g.y,
       &man->managed_g.width, &man->managed_g.height);
-    scr = FSCREEN_XYPOS;
   }
   else
   {
@@ -930,7 +934,6 @@ void create_manager_window (int man_id)
       NULL, FSCREEN_GLOBAL,
       &man->managed_g.x, &man->managed_g.y,
       &man->managed_g.width, &man->managed_g.height);
-    scr = FSCREEN_GLOBAL;
   }
 
   man->theWindow = XCreateWindow(theDisplay, theRoot, sizehints.x, sizehints.y,
@@ -938,7 +941,7 @@ void create_manager_window (int man_id)
 				 0, Pdepth, InputOutput, Pvisual, winattrmask,
 				 &winattr);
   /* hack to prevent mapping on wrong screen with StartsOnScreen */
-  FScreenMangleScreenIntoUSPosHints(scr, &sizehints);
+  FScreenMangleScreenIntoUSPosHints(FSCREEN_XYPOS, &sizehints);
   XSetWMNormalHints(theDisplay, man->theWindow, &sizehints);
   FShapeSelectInput (theDisplay, man->theWindow, FShapeNotifyMask);
   if (globals.transient)
@@ -964,6 +967,8 @@ void create_manager_window (int man_id)
   for (i = 0; i < NUM_CONTEXTS; i++) {
 	  if (man->colorsets[i] != -2)
 	  {
+		  gcmask = GCForeground;
+		  gcval.foreground = man->backcolor[i];
 		  man->backContext[i] =
 			  fvwmlib_XCreateGC(
 				  theDisplay, man->theWindow, gcmask, &gcval);
@@ -974,6 +979,8 @@ void create_manager_window (int man_id)
 			  line_style, cap_style,
 			  join_style);
 	  }
+	  gcmask = GCForeground;
+	  gcval.foreground = man->forecolor[i];
 	  man->hiContext[i] = fvwmlib_XCreateGC(
 		  theDisplay, man->theWindow, gcmask, &gcval);
 	  if (man->FButtonFont->font != NULL)
