@@ -2,11 +2,7 @@
 /* This module, and the entire ModuleDebugger program, and the concept for
  * interfacing this module to the Window Manager, are all original work
  * by Robert Nation
- *
- * Copyright 1994, Robert Nation. No guarantees or warantees or anything
- * are provided or implied in any way whatsoever. Use this program at your
- * own risk. Permission to use this program for any purpose is given,
- * as long as the copyright is kept intact. */
+ */
 
 /* This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,8 +15,7 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * along with this program; if not, see: <http://www.gnu.org/licenses/>
  */
 
 #include "config.h"
@@ -45,7 +40,6 @@
 
 #include "libs/Module.h"
 #include "libs/fvwmlib.h"
-#include "libs/FScreen.h"
 #include "libs/FShape.h"
 #include "libs/FRenderInit.h"
 #include "libs/Colorset.h"
@@ -68,6 +62,8 @@ PagerStringList *NewPagerStringItem(PagerStringList *last, int desk);
 extern FlocaleFont *FwindowFont;
 extern Pixmap default_pixmap;
 
+struct fpmonitors		 fp_monitor_q;
+
 /*
  *
  * Screen, font, etc info
@@ -119,6 +115,7 @@ char *BalloonFore = NULL;
 char *BalloonFont = NULL;
 char *BalloonBorderColor = NULL;
 char *BalloonFormatString = NULL;
+char *monitor_to_track = NULL;
 int BalloonBorderWidth = 1;
 int BalloonYOffset = 3;
 Window BalloonView = None;
@@ -154,9 +151,67 @@ Bool do_ignore_next_button_release = False;
 Bool error_occured = False;
 Bool Swallowed = False;
 
-static void SetDeskLabel(int desk, const char *label);
+static void extract_monitor_config(struct fpmonitor *, char *);
+static void SetDeskLabel(struct fpmonitor *, int desk, const char *label);
 static RETSIGTYPE TerminateHandler(int);
 void ExitPager(void);
+void list_monitor_focus(unsigned long *);
+
+struct fpmonitor *
+fpmonitor_get_current(void)
+{
+	struct fpmonitor *m;
+
+	TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+		if (m->is_current)
+			return (m);
+	}
+	return (NULL);
+}
+
+struct fpmonitor *
+fpmonitor_by_name(const char *name)
+{
+	struct fpmonitor	*m;
+
+	TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+		if (strcmp(m->name, name) == 0)
+			return (m);
+	}
+	return (NULL);
+}
+
+struct fpmonitor *
+fpmonitor_by_output(int output)
+{
+	struct fpmonitor	*m;
+
+	TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+		if (m->output == output)
+			return (m);
+	}
+	return (NULL);
+}
+
+
+
+struct fpmonitor *
+fpmonitor_this(void)
+{
+	struct fpmonitor 	*m;
+
+	if (monitor_to_track == NULL)
+		return (fpmonitor_get_current());
+
+	TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+		if (monitor_to_track != NULL &&
+		    strcmp(m->name, monitor_to_track) == 0) {
+			return (m);
+		}
+	}
+
+	return (fpmonitor_get_current());
+}
 
 /*
  *
@@ -176,16 +231,22 @@ int main(int argc, char **argv)
 
   FlocaleInit(LC_CTYPE, "", "", "FvwmPager");
 
+  /* Tell the FEvent module an event type that is not used by fvwm. */
+  fev_init_invalid_event_type(KeymapNotify);
+
   /* Save our program  name - for error messages */
   MyName = GetFileNameFromPath(argv[0]);
 
   if(argc  < 6)
     {
-      fprintf(stderr,"%s Version %s should only be executed by fvwm!\n",MyName,
-	      VERSION);
+      fvwm_debug(__func__, "%s Version %s should only be executed by fvwm!\n",
+                 MyName,
+                 VERSION);
       exit(1);
     }
 
+  TAILQ_INIT(&fp_monitor_q);
+
 #ifdef HAVE_SIGACTION
   {
     struct sigaction  sigact;
@@ -259,7 +320,7 @@ int main(int argc, char **argv)
 	      (*s != '-' || s != argv[opt_num] || *(s+1) == 0))
 	  {
 	    free(MyName);
-	    MyName=safestrdup(argv[opt_num]);
+	    MyName = fxstrdup(argv[opt_num]);
 	    opt_num++;
 	    break;
 	  }
@@ -269,13 +330,13 @@ int main(int argc, char **argv)
 
   if (argc < opt_num + 1)
     {
-      desk1 = Scr.CurrentDesk;
-      desk2 = Scr.CurrentDesk;
+      desk1 = 0; //Scr.CurrentDesk;
+      desk2 = 0; //Scr.CurrentDesk;
     }
   else if (StrEquals(argv[opt_num], "*"))
     {
-      desk1 = Scr.CurrentDesk;
-      desk2 = Scr.CurrentDesk;
+      desk1 = 0; //Scr.CurrentDesk;
+      desk2 = 0; //Scr.CurrentDesk;
       fAlwaysCurrentDesk = 1;
     }
   else
@@ -294,8 +355,7 @@ int main(int argc, char **argv)
     }
   ndesks = desk2 - desk1 + 1;
 
-  Desks = (DeskInfo *)safemalloc(ndesks*sizeof(DeskInfo));
-  memset(Desks, 0, ndesks * sizeof(DeskInfo));
+  Desks = fxcalloc(1, ndesks*sizeof(DeskInfo));
   for(i=0;i<ndesks;i++)
     {
       sprintf(line,"Desk %d",i+desk1);
@@ -308,8 +368,8 @@ int main(int argc, char **argv)
   /* Initialize X connection */
   if (!(dpy = XOpenDisplay(display_name)))
     {
-      fprintf(stderr,"%s: can't open display %s", MyName,
-	      XDisplayName(display_name));
+      fvwm_debug(__func__, "%s: can't open display %s", MyName,
+                 XDisplayName(display_name));
       exit (1);
     }
 
@@ -344,7 +404,8 @@ int main(int argc, char **argv)
   SetMessageMask(fd,
 		 MX_VISIBLE_ICON_NAME|
 		 MX_PROPERTY_CHANGE|
-		 MX_REPLY);
+		 MX_MONITOR_FOCUS);
+
   ParseOptions();
   if (is_transient)
   {
@@ -362,37 +423,37 @@ int main(int argc, char **argv)
   }
 
   if (PagerFore == NULL)
-    PagerFore = safestrdup("black");
+    PagerFore = fxstrdup("black");
 
   if (PagerBack == NULL)
-    PagerBack = safestrdup("white");
+    PagerBack = fxstrdup("white");
 
   if (HilightC == NULL)
-    HilightC = safestrdup(PagerFore);
+    HilightC = fxstrdup(PagerFore);
 
   if (WindowLabelFormat == NULL)
-    WindowLabelFormat = safestrdup("%i");
+    WindowLabelFormat = fxstrdup("%i");
 
   if ((HilightC == NULL) && (HilightPixmap == NULL))
     HilightDesks = 0;
 
   if (BalloonBorderColor == NULL)
-    BalloonBorderColor = safestrdup("black");
+    BalloonBorderColor = fxstrdup("black");
 
   if (BalloonTypeString == NULL)
-    BalloonTypeString = safestrdup("%i");
+    BalloonTypeString = fxstrdup("%i");
 
   if (BalloonFormatString == NULL)
-    BalloonFormatString = safestrdup("%i");
-
-  /* open a pager window */
-  initialize_pager();
+    BalloonFormatString = fxstrdup("%i");
 
   /* Create a list of all windows */
   /* Request a list of all windows,
    * wait for ConfigureWindow packets */
   SendInfo(fd,"Send_WindowList",0);
 
+  /* open a pager window */
+  initialize_pager();
+
   if (is_transient)
   {
     Bool is_pointer_grabbed = False;
@@ -423,9 +484,9 @@ int main(int argc, char **argv)
     if (!is_pointer_grabbed)
     {
       XBell(dpy, 0);
-      fprintf(stderr,
-	      "%s: could not grab pointer in transient mode. exiting.\n",
-	      MyName);
+      fvwm_debug(__func__,
+                 "%s: could not grab pointer in transient mode. exiting.\n",
+                 MyName);
       exit(1);
     }
 
@@ -439,7 +500,7 @@ int main(int argc, char **argv)
 #ifdef FVWM_DEBUG_MSGS
   if ( isTerminated )
   {
-    fprintf(stderr, "%s: Received signal: exiting...\n", MyName);
+    fvwm_debug(__func__, "%s: Received signal: exiting...\n", MyName);
   }
 #endif
 
@@ -547,6 +608,9 @@ void process_message( FvwmPacket* packet )
     case MX_PROPERTY_CHANGE:
       list_property_change(body);
       break;
+    case MX_MONITOR_FOCUS:
+      list_monitor_focus(body);
+      break;
     case MX_REPLY:
 	    list_reply(body);
 	    break;
@@ -588,7 +652,8 @@ void handle_config_win_package(PagerWindow *t,
 	if (t->w != None && t->w != cfgpacket->w)
 	{
 		/* Should never happen */
-		fprintf(stderr,"%s: Error: Internal window list corrupt\n",MyName);
+		fvwm_debug(__func__,
+			   "%s: Error: Internal window list corrupt\n",MyName);
 		/* might be a good idea to exit here */
 		return;
 	}
@@ -598,6 +663,7 @@ void handle_config_win_package(PagerWindow *t,
 	t->frame_y = cfgpacket->frame_y;
 	t->frame_width = cfgpacket->frame_width;
 	t->frame_height = cfgpacket->frame_height;
+	t->m = fpmonitor_by_output(cfgpacket->monitor_id);
 
 	t->desk = cfgpacket->desk;
 
@@ -657,12 +723,20 @@ void list_add(unsigned long *body)
 	PagerWindow *t,**prev;
 	int i=0;
 	struct ConfigWinPacket  *cfgpacket = (void *) body;
+	struct fpmonitor *newm;
 
 	t = Start;
 	prev = &Start;
 
-	while(t!= NULL)
+	newm = fpmonitor_by_output((int)cfgpacket->monitor_id);
+
+	if (newm == NULL)
+		newm = fpmonitor_get_current();
+
+	while(t != NULL)
 	{
+		//t->m = newm;
+		fprintf(stderr, "WINDOW %s: mon: %s\n", t->window_name, t->m->name);
 		if (t->w == cfgpacket->w)
 		{
 			/* it's already there, do nothing */
@@ -672,8 +746,7 @@ void list_add(unsigned long *body)
 		t = t->next;
 		i++;
 	}
-	*prev = (PagerWindow *)safemalloc(sizeof(PagerWindow));
-	memset(*prev, 0, sizeof(PagerWindow));
+	*prev = fxcalloc(1, sizeof(PagerWindow));
 	handle_config_win_package(*prev, cfgpacket);
 	AddNewWindow(*prev);
 
@@ -692,11 +765,30 @@ void list_configure(unsigned long *body)
   Window target_w;
   struct ConfigWinPacket  *cfgpacket = (void *) body;
   Bool is_new_desk;
+  struct fpmonitor *newm;
 
   target_w = cfgpacket->w;
   t = Start;
-  while((t!= NULL)&&(t->w != target_w))
+  newm = fpmonitor_by_output((int)cfgpacket->monitor_id);
+
+  if (newm == NULL) {
+	  fvwm_debug(__func__, "monitor was null with ID: %d\n",
+                     (int)cfgpacket->monitor_id);
+	  fvwm_debug(__func__, "using current montior\n");
+
+	  newm = fpmonitor_get_current();
+  }
+
+  /* FIXME: need to handle things when DesktopConfiguration is global --
+   * and/or when the Monitor string insn't set in FvwmPager.
+   */
+
+  while( (t != NULL) && (t->w != target_w))
   {
+	  if (t->m != newm) {
+		  t = t->next;
+		  continue;
+	  }
     t = t->next;
   }
   if(t== NULL)
@@ -759,6 +851,28 @@ void list_destroy(unsigned long *body)
     }
 }
 
+void list_monitor_focus(unsigned long *body)
+{
+	struct fpmonitor	*m, *mcur, *mthis;
+	char	*mon_name = (char *)(&body[3]);
+
+	mcur = fpmonitor_get_current();
+
+	if (mcur == NULL || mon_name == NULL)
+		return;
+
+	mthis = fpmonitor_by_name(mon_name);
+
+	if (mcur == mthis)
+		return;
+
+	mcur->is_current = 0;
+	TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+		if (m == mthis)
+			m->is_current = 1;
+	}
+}
+
 /*
  *
  *  Procedure:
@@ -814,22 +928,31 @@ void list_focus(unsigned long *body)
  */
 void list_new_page(unsigned long *body)
 {
-  Scr.Vx = body[0];
-  Scr.Vy = body[1];
-  if (Scr.CurrentDesk != body[2])
+	int mon_num = body[7];
+	struct fpmonitor *m;
+
+	m = fpmonitor_by_output(mon_num);
+	fvwm_debug(__func__, "%s: using monitor: %s\n", __func__, m->name);
+
+	if (monitor_to_track != NULL && strcmp(m->name, monitor_to_track) != 0)
+		return;
+
+  m->virtual_scr.Vx = body[0];
+  m->virtual_scr.Vy = body[1];
+  if (m->virtual_scr.CurrentDesk != body[2])
   {
       /* first handle the new desk */
       body[0] = body[2];
       list_new_desk(body);
   }
-  if (Scr.VxPages != body[5] || Scr.VyPages != body[6])
+  if (m->virtual_scr.VxPages != body[5] || m->virtual_scr.VyPages != body[6])
   {
-    Scr.VxPages = body[5];
-    Scr.VyPages = body[6];
-    Scr.VWidth = Scr.VxPages * Scr.MyDisplayWidth;
-    Scr.VHeight = Scr.VyPages * Scr.MyDisplayHeight;
-    Scr.VxMax = Scr.VWidth - Scr.MyDisplayWidth;
-    Scr.VyMax = Scr.VHeight - Scr.MyDisplayHeight;
+    m->virtual_scr.VxPages = body[5];
+    m->virtual_scr.VyPages = body[6];
+    m->virtual_scr.VWidth = m->virtual_scr.VxPages * m->virtual_scr.MyDisplayWidth;
+    m->virtual_scr.VHeight = m->virtual_scr.VyPages * m->virtual_scr.MyDisplayHeight;
+    m->virtual_scr.VxMax = m->virtual_scr.VWidth - m->virtual_scr.MyDisplayWidth;
+    m->virtual_scr.VyMax = m->virtual_scr.VHeight - m->virtual_scr.MyDisplayHeight;
     ReConfigure();
   }
   MovePage(False);
@@ -849,23 +972,29 @@ void list_new_desk(unsigned long *body)
   int change_cs = -1;
   int change_ballooncs = -1;
   int change_highcs = -1;
+  struct fpmonitor *m;
 
-  oldDesk = Scr.CurrentDesk;
-  Scr.CurrentDesk = (long)body[0];
-  if (fAlwaysCurrentDesk && oldDesk != Scr.CurrentDesk)
+  m = fpmonitor_this();
+
+  if (monitor_to_track != NULL && strcmp(m->name, monitor_to_track) != 0)
+	  return;
+
+  oldDesk = m->virtual_scr.CurrentDesk;
+  m->virtual_scr.CurrentDesk = (long)body[0];
+  if (fAlwaysCurrentDesk && oldDesk != m->virtual_scr.CurrentDesk)
   {
     PagerWindow *t;
     PagerStringList *item;
     char line[100];
 
-    desk1 = Scr.CurrentDesk;
-    desk2 = Scr.CurrentDesk;
+    desk1 = m->virtual_scr.CurrentDesk;
+    desk2 = m->virtual_scr.CurrentDesk;
     for (t = Start; t != NULL; t = t->next)
     {
-      if (t->desk == oldDesk || t->desk == Scr.CurrentDesk)
+      if (t->desk == oldDesk || t->desk == m->virtual_scr.CurrentDesk)
 	ChangeDeskForWindow(t, t->desk);
     }
-    item = FindDeskStrings(Scr.CurrentDesk);
+    item = FindDeskStrings(m->virtual_scr.CurrentDesk);
     if (Desks[0].label != NULL)
     {
       free(Desks[0].label);
@@ -976,14 +1105,14 @@ void list_new_desk(unsigned long *body)
     char *name;
     char line[100];
 
-    i = Scr.CurrentDesk - desk1;
+    i = m->virtual_scr.CurrentDesk - desk1;
     if (i >= 0 && i < ndesks && Desks[i].label != NULL)
     {
       name = Desks[i].label;
     }
     else
     {
-      sprintf(line, "Desk %d", Scr.CurrentDesk);
+      sprintf(line, "Desk %d", m->virtual_scr.CurrentDesk);
       name = &(line[0]);
     }
     XStoreName(dpy, Scr.Pager_w, name);
@@ -992,7 +1121,7 @@ void list_new_desk(unsigned long *body)
 
   MovePage(True);
   DrawGrid(oldDesk - desk1, 1, None, NULL);
-  DrawGrid(Scr.CurrentDesk - desk1, 1, None, NULL);
+  DrawGrid(m->virtual_scr.CurrentDesk - desk1, 1, None, NULL);
   MoveStickyWindow(False, True);
 /*
   Hilight(FocusWin,False);
@@ -1157,11 +1286,20 @@ void list_window_name(unsigned long *body,unsigned long type)
 {
   PagerWindow *t;
   Window target_w;
+  struct fpmonitor *m = fpmonitor_this();
+
+
+  if (monitor_to_track != NULL && strcmp(m->name, monitor_to_track) != 0)
+	  return;
 
   target_w = body[0];
   t = Start;
   while((t!= NULL)&&(t->w != target_w))
     {
+	    if (m != NULL && t->m != m) {
+		    t = t->next;
+		    continue;
+	    }
       t = t->next;
     }
   if(t!= NULL)
@@ -1188,7 +1326,7 @@ void list_window_name(unsigned long *body,unsigned long type)
 	   && !(MiniIcons && t->mini_icon.picture)) {
 	if (t->PagerView)
 	  XClearArea(dpy, t->PagerView, 0, 0, 0, 0, True);
-	if (t->IconView);
+	if (t->IconView)
 	  XClearArea(dpy, t->IconView, 0, 0, 0, 0, True);
       }
       if (ShowBalloons && BalloonView)
@@ -1235,7 +1373,7 @@ void list_icon_name(unsigned long *body)
 	   && !(MiniIcons && t->mini_icon.picture)) {
 	if (t->PagerView)
 	  XClearArea(dpy, t->PagerView, 0, 0, 0, 0, True);
-	if (t->IconView);
+	if (t->IconView)
 	  XClearArea(dpy, t->IconView, 0, 0, 0, 0, True);
       }
     }
@@ -1245,10 +1383,8 @@ void list_icon_name(unsigned long *body)
 void list_mini_icon(unsigned long *body)
 {
   PagerWindow   *t;
-  Window target_w;
   MiniIconPacket *mip = (MiniIconPacket *) body;
 
-  target_w = body[0];
   t = Start;
   while (t && (t->w != mip->w))
     t = t->next;
@@ -1264,7 +1400,7 @@ void list_mini_icon(unsigned long *body)
     if (MiniIcons && t->mini_icon.picture) {
       if (t->PagerView)
 	XClearArea(dpy, t->PagerView, 0, 0, 0, 0, True);
-      if (t->IconView);
+      if (t->IconView)
 	XClearArea(dpy, t->IconView, 0, 0, 0, 0, True);
     }
   }
@@ -1277,7 +1413,7 @@ void list_restack(unsigned long *body, unsigned long length)
   Window *wins;
   int i, j, d;
 
-  wins = (Window *) safemalloc (length * sizeof (Window));
+  wins = fxmalloc(length * sizeof (Window));
   /* first restack in the icon view */
   j = 0;
   for (i = 0; i < (length - FvwmPacketHeaderSize); i += 3)
@@ -1358,45 +1494,108 @@ void list_end(void)
     XFree((char *)children);
 }
 
+void extract_monitor_config(struct fpmonitor *m, char *tline)
+{
+    int  output, mdw, mdh, vx, vy, vxmax, vymax, iscur;
+    int  x, y, w, h;
+
+    sscanf(tline, "%d %d %d %d %d %d %d %d %d %d %d %d",
+        &output, &iscur, &mdw, &mdh, &vx, &vy, &vxmax, &vymax,
+        &x, &y, &w, &h);
+    m->x = x;
+    m->y = y;
+    m->w = w;
+    m->h = h;
+    m->output = output;
+    m->is_current = iscur;
+    m->virtual_scr.MyDisplayWidth = mdw;
+    m->virtual_scr.MyDisplayHeight = mdh;
+    m->virtual_scr.Vx = vx;
+    m->virtual_scr.Vy = vy;
+    m->virtual_scr.VxMax = vxmax;
+    m->virtual_scr.VyMax = vymax;
+
+}
+
 
 void list_config_info(unsigned long *body)
 {
-  char *tline, *token;
-  int color;
+	struct fpmonitor *m, *m2;
+	char *tline, *token;
+	int color;
 
-  tline = (char*)&(body[3]);
-  token = PeekToken(tline, &tline);
-  if (StrEquals(token, "Colorset"))
-  {
-    color = LoadColorset(tline);
-    change_colorset(color);
-  }
-  else if (StrEquals(token, XINERAMA_CONFIG_STRING))
-  {
-    FScreenConfigureModule(tline);
-  }
-  else if (StrEquals(token, "DesktopName"))
-  {
-    int val;
-    if (GetIntegerArguments(tline, &tline, &val, 1) > 0)
-    {
-      SetDeskLabel(val, (const char *)tline);
-    }
-    else
-    {
-      return;
-    }
-    if (fAlwaysCurrentDesk)
-    {
-      if (Scr.CurrentDesk == val)
-	val = 0;
-    }
-    else if ((val >= desk1) && (val <=desk2))
-    {
-      val = val - desk1;
-    }
-    DrawGrid(val, True, None, NULL);
-  }
+	tline = (char*)&(body[3]);
+	token = PeekToken(tline, &tline);
+	if (StrEquals(token, "Colorset"))
+	{
+		color = LoadColorset(tline);
+		change_colorset(color);
+	}
+	else if (StrEquals(token, "DesktopName"))
+	{
+		int val;
+		if (GetIntegerArguments(tline, &tline, &val, 1) > 0)
+		{
+			TAILQ_FOREACH(m, &fp_monitor_q, entry)
+				SetDeskLabel(m, val, (const char *)tline);
+		}
+		else
+		{
+			return;
+		}
+		if (fAlwaysCurrentDesk)
+		{
+			TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+				if (m->virtual_scr.CurrentDesk == val)
+					val = 0;
+			}
+		}
+		else if ((val >= desk1) && (val <=desk2))
+		{
+			val = val - desk1;
+		}
+		DrawGrid(val, True, None, NULL);
+	} else if (StrEquals(token, "Monitor")) {
+		char	*mname;
+		int		 updated = 0;
+
+		tline = GetNextToken(tline, &mname);
+
+		TAILQ_FOREACH(m2, &fp_monitor_q, entry) {
+			updated = 0;
+			if (strcmp(m2->name, mname) == 0) {
+				extract_monitor_config(m2,tline);
+				updated = 1;
+			}
+		}
+
+		if (updated)
+			return;
+
+		m = fxcalloc(1, sizeof(*m));
+
+		m->name = fxstrdup(mname);
+		extract_monitor_config(m,tline);
+		TAILQ_INSERT_TAIL(&fp_monitor_q, m, entry);
+	} else if (StrEquals(token, "DesktopSize")) {
+		int dx, dy;
+		struct fpmonitor *m;
+
+		sscanf(tline, "%d %d", &dx, &dy);
+
+		TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+			m->virtual_scr.VxMax = dx * m->virtual_scr.MyDisplayWidth - m->virtual_scr.MyDisplayWidth;
+			m->virtual_scr.VyMax = dy * m->virtual_scr.MyDisplayHeight - m->virtual_scr.MyDisplayHeight;
+			if (m->virtual_scr.VxMax < 0)
+				m->virtual_scr.VxMax = 0;
+			if (m->virtual_scr.VyMax < 0)
+				m->virtual_scr.VyMax = 0;
+			m->virtual_scr.VWidth = m->virtual_scr.VxMax + m->virtual_scr.MyDisplayWidth;
+			m->virtual_scr.VHeight = m->virtual_scr.VyMax + m->virtual_scr.MyDisplayHeight;
+			m->virtual_scr.VxPages = m->virtual_scr.VWidth / m->virtual_scr.MyDisplayWidth;
+			m->virtual_scr.VyPages = m->virtual_scr.VHeight / m->virtual_scr.MyDisplayHeight;
+		}
+	}
 }
 
 void list_property_change(unsigned long *body)
@@ -1434,7 +1633,6 @@ void list_reply(unsigned long *body)
 int My_XNextEvent(Display *dpy, XEvent *event)
 {
   fd_set in_fdset;
-  static int miss_counter = 0;
 
   if(FPending(dpy))
   {
@@ -1453,14 +1651,8 @@ int My_XNextEvent(Display *dpy, XEvent *event)
       if(FPending(dpy))
 	{
 	  FNextEvent(dpy,event);
-	  miss_counter = 0;
 	  return 1;
 	}
-	miss_counter++;
-#ifdef WORRY_ABOUT_MISSED_XEVENTS
-	if(miss_counter > 100)
-	  DeadPipe(0);
-#endif
     }
 
   if(FD_ISSET(fd[1], &in_fdset))
@@ -1488,13 +1680,14 @@ static void ParseColorset(char *arg1, char *arg2, void *offset_deskinfo,
   int desk;
   unsigned long colorset_offset = (unsigned long)offset_deskinfo;
   unsigned long item_colorset_offset = (unsigned long)offset_item;
+  struct fpmonitor *mon = fpmonitor_this();
 
   sscanf(arg2, "%d", &colorset);
   AllocColorset(colorset);
   if (StrEquals(arg1, "*"))
   {
     all_desks = True;
-    desk = Scr.CurrentDesk;
+    desk = 0; //Scr.CurrentDesk;
   }
   else
   {
@@ -1523,7 +1716,7 @@ static void ParseColorset(char *arg1, char *arg2, void *offset_deskinfo,
 	*(int *)(((char *)item) + item_colorset_offset) = colorset;
       }
     }
-    if (desk == Scr.CurrentDesk || all_desks)
+    if (desk == mon->virtual_scr.CurrentDesk || all_desks)
     {
       *(int *)(((char *)&Desks[0]) + colorset_offset) = colorset;
     }
@@ -1543,7 +1736,7 @@ static void ParseColorset(char *arg1, char *arg2, void *offset_deskinfo,
   return;
 }
 
-static void SetDeskLabel(int desk, const char *label)
+static void SetDeskLabel(struct fpmonitor *m, int desk, const char *label)
 {
   PagerStringList *item;
 
@@ -1566,7 +1759,7 @@ static void SetDeskLabel(int desk, const char *label)
       item = NewPagerStringItem(item, desk);
       CopyString(&(item->label), label);
     }
-    if (desk == Scr.CurrentDesk)
+    if (desk == m->virtual_scr.CurrentDesk)
     {
       free(Desks[0].label);
       CopyString(&Desks[0].label, label);
@@ -1595,9 +1788,6 @@ void ParseOptions(void)
   Scr.Hilite = NULL;
   Scr.VScale = 32;
 
-  Scr.MyDisplayWidth = DisplayWidth(dpy, Scr.screen);
-  Scr.MyDisplayHeight = DisplayHeight(dpy, Scr.screen);
-
   fpa.mask = 0;
   if (Pdepth <= 8)
   {
@@ -1615,19 +1805,17 @@ void ParseOptions(void)
     char *token;
     char *next;
     Bool MoveThresholdSetForModule = False;
+    struct fpmonitor	*m = NULL, *m2 = NULL;
 
     arg1 = arg2 = NULL;
 
     token = PeekToken(tline, &next);
+
     if (StrEquals(token, "Colorset"))
     {
       LoadColorset(next);
       continue;
     }
-    else if (StrEquals(token, XINERAMA_CONFIG_STRING))
-    {
-      FScreenConfigureModule(next);
-    }
     else if (StrEquals(token, "DesktopSize"))
     {
       token = PeekToken(next, &next);
@@ -1671,10 +1859,36 @@ void ParseOptions(void)
       int val;
       if (GetIntegerArguments(next, &next, &val, 1) > 0)
       {
-	SetDeskLabel(val, (const char *)next);
+	      TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+		      SetDeskLabel(m, val, (const char *)next);
+	      }
       }
       continue;
     }
+    else if (StrEquals(token, "Monitor")) {
+	    char	*mname;
+	    int		 updated = 0;
+
+	    next = GetNextToken(next, &mname);
+
+	    TAILQ_FOREACH(m2, &fp_monitor_q, entry) {
+		    updated = 0;
+		    if (strcmp(m2->name, mname) == 0) {
+			    extract_monitor_config(m2, next);
+			    updated = 1;
+		    }
+	    }
+
+	    if (updated)
+		    continue;
+
+	    m = fxcalloc(1, sizeof(*m));
+
+	    m->name = fxstrdup(mname);
+	    extract_monitor_config(m, next);
+	    TAILQ_INSERT_TAIL(&fp_monitor_q, m, entry);
+	    continue;
+    }
 
     tline2 = GetModuleResource(tline, &resource, MyName);
     if (!resource)
@@ -1682,17 +1896,36 @@ void ParseOptions(void)
     tline2 = GetNextToken(tline2, &arg1);
     if (!arg1)
     {
-      arg1 = (char *)safemalloc(1);
+      arg1 = fxmalloc(1);
       arg1[0] = 0;
     }
     tline2 = GetNextToken(tline2, &arg2);
     if (!arg2)
     {
-      arg2 = (char *)safemalloc(1);
+      arg2 = fxmalloc(1);
       arg2[0] = 0;
     }
 
-    if(StrEquals(resource,"Colorset"))
+    if (StrEquals(resource, "Monitor")) {
+	    free(monitor_to_track);
+	    if (next == NULL) {
+		    fvwm_debug(__func__, "FvwmPager: no monitor name given "
+				    "using current monitor\n");
+		    /* m already set... */
+		    monitor_to_track = fxstrdup(m->name);
+		    continue;
+	    }
+	    if ((m = fpmonitor_by_name(next)) == NULL) {
+		    fvwm_debug(__func__, "FvwmPager: monitor '%s' not found "
+                               "using current monitor", next);
+		    m = fpmonitor_get_current();
+		    monitor_to_track = fxstrdup(m->name);
+		    continue;
+	    }
+	    fvwm_debug(__func__, "Assigning monitor: %s\n", m->name);
+	    monitor_to_track = fxstrdup(m->name);
+    }
+    else if(StrEquals(resource,"Colorset"))
     {
       ParseColorset(arg1, arg2,
 		    &(((DeskInfo *)(NULL))->colorset),
@@ -1784,14 +2017,15 @@ void ParseOptions(void)
     {
       if (StrEquals(arg1, "*"))
       {
-	desk = Scr.CurrentDesk;
+	desk = 0; //Scr.CurrentDesk;
       }
       else
       {
 	desk = desk1;
 	sscanf(arg1,"%d",&desk);
       }
-      SetDeskLabel(desk, (const char *)arg2);
+      TAILQ_FOREACH(m, &fp_monitor_q, entry)
+	      SetDeskLabel(m, desk, (const char *)arg2);
     }
     else if (StrEquals(resource, "Font"))
     {
@@ -1830,7 +2064,7 @@ void ParseOptions(void)
     {
       if (StrEquals(arg1, "*"))
       {
-	desk = Scr.CurrentDesk;
+	desk = 0; //Scr.CurrentDesk;
       }
       else
       {
@@ -1858,7 +2092,9 @@ void ParseOptions(void)
 	  item = NewPagerStringItem(item, desk);
 	  CopyString(&(item->Dcolor), arg2);
 	}
-	if (desk == Scr.CurrentDesk)
+	m = fpmonitor_this();
+
+	if (desk == m->virtual_scr.CurrentDesk)
 	{
 	  free(Desks[0].Dcolor);
 	  CopyString(&Desks[0].Dcolor, arg2);
@@ -1874,7 +2110,7 @@ void ParseOptions(void)
     {
       if (StrEquals(arg1, "*"))
       {
-	desk = Scr.CurrentDesk;
+	desk = 0; //Scr.CurrentDesk;
       }
       else
       {
@@ -1904,7 +2140,8 @@ void ParseOptions(void)
 	  item->bgPixmap = PCacheFvwmPicture(
 		  dpy, Scr.Pager_w, ImagePath, arg2, fpa);
 	}
-	if (desk == Scr.CurrentDesk)
+	m = fpmonitor_this();
+	if (desk == m->virtual_scr.CurrentDesk)
 	{
 	  if (Desks[0].bgPixmap != NULL)
 	  {
@@ -2163,10 +2400,10 @@ void ParseOptions(void)
       sscanf(arg1, "%d", &BalloonYOffset);
       if (BalloonYOffset == 0)
       {
-	fprintf(stderr,
-		"%s: Warning:"
-		" you're not allowed BalloonYOffset 0; defaulting to +3\n",
-		MyName);
+	fvwm_debug(__func__,
+                   "%s: Warning:"
+                   " you're not allowed BalloonYOffset 0; defaulting to +3\n",
+                   MyName);
 	/* we don't allow yoffset of 0 because it allows direct transit from
 	 * pager window to balloon window, setting up a LeaveNotify/EnterNotify
 	 * event loop */
@@ -2185,18 +2422,26 @@ void ParseOptions(void)
     free(arg2);
   }
 
-  Scr.VxMax = dx * Scr.MyDisplayWidth - Scr.MyDisplayWidth;
-  Scr.VyMax = dy * Scr.MyDisplayHeight - Scr.MyDisplayHeight;
-  if (Scr.VxMax < 0)
-    Scr.VxMax = 0;
-  if (Scr.VyMax < 0)
-    Scr.VyMax = 0;
-  Scr.VWidth = Scr.VxMax + Scr.MyDisplayWidth;
-  Scr.VHeight = Scr.VyMax + Scr.MyDisplayHeight;
-  Scr.VxPages = Scr.VWidth / Scr.MyDisplayWidth;
-  Scr.VyPages = Scr.VHeight / Scr.MyDisplayHeight;
-  Scr.Vx = 0;
-  Scr.Vy = 0;
+  struct fpmonitor *m;
+  TAILQ_FOREACH(m, &fp_monitor_q, entry) {
+	  m->virtual_scr.VxMax = dx * m->virtual_scr.MyDisplayWidth - m->virtual_scr.MyDisplayWidth;
+	  m->virtual_scr.VyMax = dy * m->virtual_scr.MyDisplayHeight - m->virtual_scr.MyDisplayHeight;
+	  if (m->virtual_scr.VxMax < 0)
+		  m->virtual_scr.VxMax = 0;
+	  if (m->virtual_scr.VyMax < 0)
+		  m->virtual_scr.VyMax = 0;
+	  m->virtual_scr.VWidth = m->virtual_scr.VxMax + m->virtual_scr.MyDisplayWidth;
+	  m->virtual_scr.VHeight = m->virtual_scr.VyMax + m->virtual_scr.MyDisplayHeight;
+	  m->virtual_scr.VxPages = m->virtual_scr.VWidth / m->virtual_scr.MyDisplayWidth;
+	  m->virtual_scr.VyPages = m->virtual_scr.VHeight / m->virtual_scr.MyDisplayHeight;
+
+	  fvwm_debug(__func__,
+			  "%s: VxMax: %d, VyMax: %d, VWidth: %d, Vheight: %d, "
+			  "VxPages: %d, VyPages: %d\n", m->name,
+			  m->virtual_scr.VxMax, m->virtual_scr.VyMax, m->virtual_scr.VWidth,
+			  m->virtual_scr.VHeight, m->virtual_scr.VxPages, m->virtual_scr.VyPages);
+
+  }
 
   return;
 }
@@ -2221,8 +2466,7 @@ PagerStringList *NewPagerStringItem(PagerStringList *last, int desk)
 {
   PagerStringList *newitem;
 
-  newitem = (PagerStringList *)safemalloc(sizeof(PagerStringList));
-  memset(newitem, 0, sizeof(PagerStringList));
+  newitem = fxcalloc(1, sizeof(PagerStringList));
   last->next = newitem;
   newitem->colorset = -1;
   newitem->highcolorset = -1;
